import ims
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator
from matplotlib.colors import CenteredNorm
from sklearn.cross_decomposition import PLSRegression
from sklearn.metrics import mean_squared_error, r2_score


class PLSR:
    """
    Applies a scikit-learn PLSRegression to GC-IMS data and provides prebuilt
    plots as well as a feature selection via variable importance in projection (VIP)
    scores.

    See the scikit-learn documentation for more details:
    https://scikit-learn.org/stable/modules/generated/sklearn.cross_decomposition.PLSRegression.html

    Parameters
    ----------
    dataset : ims.Dataset
        Needed for the retention and drift time coordinates in the plots.

    n_components : int, optional
        Number of components to keep, by default 2.

    kwargs : optional
        Additional key word arguments are passed on to the scikit-learn PLSRegression.

    Attributes
    ----------
    x_scores : numpy.ndarray of shape (n_samples, n_components)
        X scores.

    y_scores : numpy.ndarray of shape (n_samples, n_components)
        y scores.

    x_weights : numpy.ndarray of shape (n_features, n_components)
        The left singular vectors of the cross-covariance matrices of each iteration.

    y_weights : numpy.ndarray of shape (n_targets, n_components)
        The right singular vectors of the cross-covariance matrices of each iteration.

    x_loadings : numpy.ndarray of shape (n_features, n_components)
        The loadings of X. When scaling was applied on the dataset,
        corrects the loadings using the weights.

    y_loadings : numpy.ndarray of shape (n_targes, n_components)
        The loadings of y.

    coefficients : numpy.ndarray of shape (n_features, n_targets)
        The coefficients of the linear model.
 
    y_pred_train : numpy.ndarray
        Stores the predicted values from the training data for the plot method.
        
    rmse : float
        If y_test is set in predict method calculates root mean squared error.
        
    r2_score : float
        If y_test is set in predict method calculates R^2 score.

    Example
    -------
    >>> import ims
    >>> import pandas as pd
    >>> ds = ims.Dataset.read_mea("IMS_data")
    >>> responses = pd.read_csv("responses.csv")
    >>> ds.labels = responses
    >>> X_train, X_test, y_train, y_test = ds.train_test_split()
    >>> model = ims.PLSR(ds, n_components=5)
    >>> model.fit(X_train, y_train)
    >>> model.predict(X_test, y_test)
    >>> model.plot()
    """
    def __init__(self, dataset, n_components=2, **kwargs):
        self.dataset = dataset
        self.n_components = n_components
        self._sk_pls = PLSRegression(n_components=n_components, scale=False, **kwargs)

    def fit(self, X_train, y_train):
        """
        Fits the model with training data.

        Parameters
        ----------
        X_train : numpy.ndarray of targets (n_samples, n_features)
            Training vectors with features.

        y_train : numpy.ndarray of shape (n_samples, n_targets)
            Target vectors with response variables.

        Returns
        -------
        self
            Fitted model.
        """
        self._sk_pls.fit(X_train, y_train)
        self.x_scores, self.y_scores = self._sk_pls.transform(X_train, y_train)
        self.x_weights = self._sk_pls.x_weights_
        self.x_loadings = self._sk_pls.x_loadings_
        self.y_weights = self._sk_pls.y_weights_
        self.y_loadings = self._sk_pls.y_loadings_
        self.coefficients = self._sk_pls.coef_
        self.y_pred_train = self._sk_pls.predict(X_train)
        self.y_train = y_train
        self.vip_scores = ims.utils.vip_scores(
            self.x_weights,
            self.x_scores,
            self.y_loadings
            )
        self.selectivity_ratio = ims.utils.selectivity_ratio(
            X_train,
            self.coefficients
            )
        return self

    def predict(self, X_test, y_test=None):
        """
        Predicts responses for features of the test data.
        If y_test is set calculates root mean squared error and
        R^2 score.

        Parameters
        ----------
        X_test : numpy.ndarray of shape (n_samples, n_features)
            Features of test data.

        y_test : numpy.ndarray of shape (n_samples, n_targets), optional
            Target vector of response variables. If set calculates error metrics,
            by default None.

        Returns
        -------
        numpy.ndarray of shape (n_samples, n_targets)
            Predicted responses for test data.
        """
        y_pred = self._sk_pls.predict(X_test)
        if y_test is not None:
            self.rmse = round(mean_squared_error(y_test, y_pred, squared=False), 2)
            self.r2_score = round(r2_score(y_pred, y_test), 2)
            self.y_pred_test = y_pred
            self.y_test = y_test
        return y_pred
    
    def score(self, X_test, y_test, sample_weight=None):
        """
        Calculates R^2 score score for predicted data.

        Parameters
        ----------
        X_test : numpy.ndarray of shape (n_samples, n_features)
            Feature vectors of the test data.

        y_test : numpy.ndarray of shape (n_samples, n_targets)
            True regression responses.

        sample_weight : array-like of shape (n_samples,), optional
            Sample weights, by default None.

        Returns
        -------
        score : float
            R^2 score.
        """
        y_pred = self.predict(X_test)
        return r2_score(y_test, y_pred, sample_weight=sample_weight)
    
    
    def transform(self, X, y=None):
        """
        Apply the dimensionality reduction.

        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            Feature matrix.
        y : numpy.ndarray of shape (n_samples, n_targtets), optional
            Dependend variables, by default None

        Returns
        -------
        X_scores
        """
        return self._sk_pls.transform(X, y)


    def plot(self, width=9, height=8, annotate=False, test_only=True):
        """
        Plots predicted vs actual values and shows regression line.
        Recommended to predict with test data first.

        width : int or float, optional
            Width of the plot in inches,
            by default 9.

        height : int or float, optional
            Height of the plot in inches,
            by default 8.
            
        annotate : bool, optional
            If True annotates plot with sample names,
            by default False.
            
        test_only : bool, optional
            Ignored if annotate is set to False.
            If True and only annotates test data,
            by default True.

        Returns
        -------
        matplotlib.pyplot.axes
        """        
        _, ax = plt.subplots(figsize=(width, height))
        plt.scatter(self.y_pred_train, self.y_train, label="Train data")
        if annotate and not test_only:
            for i in range(len(self.y_train)):
                plt.annotate(
                    self.dataset[self.dataset.train_index][i].name,
                    (self.y_pred_train[i], self.y_train[i]),
                    xycoords="data"
                )

        if hasattr(self, "y_pred_test"):
            z = np.polyfit(self.y_test, self.y_pred_test, 1)
            plt.scatter(self.y_pred_test, self.y_test,
                        c="tab:orange", label="Test data")
            plt.plot(
                np.polyval(z, self.y_test),
                self.y_test,
                label=f"RMSE: {self.rmse}",
                c="tab:orange",
                linewidth=1
                )
            
            if annotate:
                for i in range(len(self.y_test)):
                    plt.annotate(
                        self.dataset[self.dataset.test_index][i].name,
                        (self.y_pred_test[i], self.y_test[i]),
                        xycoords="data"
                    )

        plt.xlabel("Predicted")
        plt.ylabel("Actual")
        plt.legend(frameon=True, fancybox=True, facecolor="white")
        return ax

    def plot_loadings(self, component=1, color_range=0.01, width=9, height=10):
        """
        Plots PLS x loadings as image with retention and drift
        time coordinates.

        Parameters
        ----------
        component : int, optional
            Component to plot, by default 1.

        color_range : float, optional
            Minimum and maximum to adjust to different scaling methods,
            by default 0.02.
 
        width : int or float, optional
            Width of the plot in inches,
            by default 9.

        height : int or float, optional
            Height of the plot in inches,
            by default 10.

        Returns
        -------
        matplotlib.pyplot.axes
        """
        loadings = self.x_loadings[:, component-1].\
            reshape(self.dataset[0].shape)
            
        ret_time = self.dataset[0].ret_time
        drift_time = self.dataset[0].drift_time

        _, ax = plt.subplots(figsize=(width, height))

        plt.imshow(
            loadings,
            cmap="RdBu_r",
            vmin=(-color_range),
            vmax=color_range,
            origin="lower",
            aspect="auto",
            extent=(min(drift_time), max(drift_time),
                    min(ret_time), max(ret_time))
            )

        plt.colorbar(label="PLS loadings")
        plt.title(f"PLS loadings of component {component}")
        plt.xlabel(self.dataset[0]._drift_time_label)
        plt.ylabel("Retention time [s]")
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        return ax
    
    def plot_coefficients(self, width=9, height=10):
        """
        Plots PLS coefficients as image with retention and drift time axis.

        Parameters
        ----------
        width : int or float, optional
            Width of the plot in inches,
            by default 9.

        height : int or float, optional
            Height of the plot in inches,
            by default 10.

        Returns
        -------
        matplotlib.pyplot.axes
        """
        coef = self.coefficients.reshape(self.dataset[0].values.shape)

        ret_time = self.dataset[0].ret_time
        drift_time = self.dataset[0].drift_time
        
        _, ax = plt.subplots(figsize=(width, height))
        
        plt.imshow(
            coef,
            cmap="RdBu_r",
            norm=CenteredNorm(0),
            origin="lower",
            aspect="auto",
            extent=(min(drift_time), max(drift_time),
                    min(ret_time), max(ret_time))
            )

        plt.colorbar(label="PLS coefficients")
        plt.title("PLS coefficients")
        plt.xlabel(self.dataset[0]._drift_time_label)
        plt.ylabel("Retention time [s]")
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        return ax
    
    def plot_vip_scores(self, threshold=None, width=9, height=10):
        """
        Plots VIP scores as image with retention and drift time axis.

        Parameters
        ----------
        width : int or float, optional
            Width of the plot in inches,
            by default 9.

        height : int or float, optional
            Height of the plot in inches,
            by default 10.

        Returns
        -------
        matplotlib.pyplot.axes

        Raises
        ------
        ValueError
            If VIP scores have not been calculated prior.
        """        
        if not hasattr(self, "vip_scores"):
            raise ValueError("Must fit data first.")

        if threshold is None:
            vip_matrix = self.vip_scores.reshape(self.dataset[0].values.shape)
        else:
            vips = np.zeros_like(self.vip_scores)
            i = np.where(self.vip_scores > threshold)
            vips[i] = self.vip_scores[i]
            vip_matrix = vips.reshape(self.dataset[0].values.shape)

        ret_time = self.dataset[0].ret_time
        drift_time = self.dataset[0].drift_time

        _, ax = plt.subplots(figsize=(width, height))

        plt.imshow(
            vip_matrix,
            cmap="RdBu_r",
            origin="lower",
            aspect="auto",
            extent=(min(drift_time), max(drift_time),
                    min(ret_time), max(ret_time))
            )

        plt.colorbar(label="VIP scores")
        plt.title(f"PLS VIP scores")
        plt.xlabel(self.dataset[0]._drift_time_label)
        plt.ylabel("Retention time [s]")
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        return ax
    
    def plot_selectivity_ratio(self, threshold=None, width=9, height=10):
        """
        Plots VIP scores as image with retention and drift time axis.
        
        Parameters
        ----------
        width : int or float, optional
            Width of the plot in inches,
            by default 9.

        height : int or float, optional
            Height of the plot in inches,
            by default 10.

        Returns
        -------
        matplotlib.pyplot.axes
        """
        if not hasattr(self, "selectivity_ratio"):
            raise ValueError("Must fit data first.")
        
        if threshold is None:
            sr_matrix = self.vip_scores.reshape(self.dataset[0].values.shape)
        else:
            sratio = np.zeros_like(self.vip_scores)
            i = np.where(self.vip_scores > threshold)
            sratio[i] = self.vip_scores[i]
            sr_matrix = sratio.reshape(self.dataset[0].values.shape)

        ret_time = self.dataset[0].ret_time
        drift_time = self.dataset[0].drift_time

        _, ax = plt.subplots(figsize=(width, height))

        plt.imshow(
            sr_matrix,
            cmap="RdBu_r",
            origin="lower",
            aspect="auto",
            extent=(min(drift_time), max(drift_time),
                    min(ret_time), max(ret_time))
            )

        plt.colorbar(label="Selectivity ratio")
        plt.title(f"PLS selectivity ratio")
        plt.xlabel(self.dataset[0]._drift_time_label)
        plt.ylabel("Retention time [s]")
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        return ax



import numpy as np
import matplotlib.pyplot as plt
from sklearn.cluster import AgglomerativeClustering
from scipy.cluster.hierarchy import dendrogram


class HCA:
    """
    Hierarchical cluster analysis with scikit-learn AgglomerativeClustering
    and scipy dendrogram.

    Parameters
    ----------
    dataset : ims.Dataset, optional
        Dataset with GC-IMS data is needed for sample and label names
        in dendrogram. If not set uses leaves as labels instead,
        by default None.

    affinity : str, optional
        Metric used to compute the linkage.
        Can be "euclidean", "l1", "l2", "manhattan" or "cosine".
        If linkage is set to "ward" only "euclidean" is accepted,
        by default "euclidean".

    linkage : str, optional
        Linkage criterion which determines which distance to use.
        "ward", "complete", "average" or "single" are accepted,
        by default "ward".

    Attributes
    ----------
    clustering : sklearn.cluster.AgglomerativeClustering
        Scikit-learn algorithm used for the clustering.
        See the original documentation for details about attributes.

    linkage_matrix : numpy.ndarray
        Clustering results encoded as linkage matrix.

    R : dict
        scipy dendrogram output as dictionary.

    Example
    -------
    >>> import ims
    >>> ds = ims.Dataset.read_mea("IMS_data")
    >>> X, _ = ds.get_xy()
    >>> hca = ims.HCA(ds, linkage="ward", affinity="euclidean")
    >>> hca.fit(X)
    >>> hca.plot_dendrogram()
    """
    def __init__(self, dataset=None, affinity="euclidean", linkage="ward",):
        self.dataset = dataset
        self.linkage = linkage
        self.affinity = affinity
        self.clustering = AgglomerativeClustering(
            distance_threshold=0,
            n_clusters=None,
            affinity=affinity,
            linkage=linkage
        )

    def fit(self, X):
        """
        Fit the model from features.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training features to cluster.
        """
        self.clustering.fit(X)
        self.linkage_matrix = self._get_linkage_matrix()
        self.R = dendrogram(self.linkage_matrix, no_plot=True)
        self.labels = self.clustering.labels_

    def plot_dendrogram(self, width=9, height=10,
                        orientation="right", **kwargs):
        """
        Plots clustering results as dendrogram.

        Parameters
        ----------
        width : int, optional
            Width of the figure in inches, by default 9

        height : int, optional
            Width of the figure in inches, by default 10

        orientation : str, optional
            Root position of the clustering tree, by default "right"

        **kwargs
            See scipy.cluster.hierarchy.dendrogram documentation
            for information about valid keyword arguments.

        Returns
        -------
        matplotlib.pyplot.axes
        """        
        _, ax = plt.subplots(figsize=(width, height))

        if self.dataset is not None:
            labels = [f"{i[0]}; {i[1]}" for i in zip(self.dataset.labels,
                                                     self.dataset.samples)]
        else:
            labels = self.R["leaves"]

        dendrogram(
            self.linkage_matrix,
            ax=ax,
            orientation=orientation,
            labels=labels,
            **kwargs
            )
        plt.xlabel(f"Distances ({self.affinity} method)")
        return ax

    def _get_linkage_matrix(self):
        """Builds linkage matrix from AgglomerativeClustering output."""
        counts = np.zeros(self.clustering.children_.shape[0])
        n_samples = len(self.clustering.labels_)
        for i, merge in enumerate(self.clustering.children_):
            current_count = 0
            for child_idx in merge:
                if child_idx < n_samples:
                    current_count += 1  # leaf node
                else:
                    current_count += counts[child_idx - n_samples]
            counts[i] = current_count

        return np.column_stack(
            [self.clustering.children_,
             self.clustering.distances_, counts]
            ).astype(float)



import ims
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator
from matplotlib.colors import CenteredNorm
import seaborn as sns
from sklearn.cross_decomposition import PLSRegression
from sklearn.metrics import (accuracy_score, precision_score,
                             recall_score)

class PLS_DA:
    """
    PLS-DA classifier built using the scikit-learn PLSRegression implementation.
    Provides prebuilt plots and feature selection via variable importance in projection (VIP)
    scores.

    See the scikit-learn documentation for more details:
    https://scikit-learn.org/stable/modules/generated/sklearn.cross_decomposition.PLSRegression.html

    Parameters
    ----------
    dataset : ims.Dataset
        Needed for the retention and drift time coordinates in the plots.

    n_components : int, optional
        Number of components to keep, by default 2.

    kwargs : optional
        Additional key word arguments are passed on to the scikit-learn PLSRegression.

    Attributes
    ----------
    x_scores : numpy.ndarray of shape (n_samples, n_components)
        X scores.

    y_scores : numpy.ndarray of shape (n_samples, n_components)
        y scores.

    x_weights : numpy.ndarray of shape (n_features, n_components)
        The left singular vectors of the cross-covariance matrices of each iteration.

    y_weights : numpy.ndarray of shape (n_targets, n_components)
        The right singular vectors of the cross-covariance matrices of each iteration.

    x_loadings : numpy.ndarray of shape (n_features, n_components)
        The loadings of X. When scaling was applied on the dataset,
        corrects the loadings using the weights.

    y_loadings : numpy.ndarray of shape (n_targes, n_components)
        The loadings of y.

    coefficients : numpy.ndarray of shape (n_features, n_targets)
        The coefficients of the linear model.
 
    y_pred_train : numpy.ndarray
        Stores the predicted values from the training data for the plot method.
        
    accuracy : float
        If y_test is set in predict method, calculates accuracy internally.
        
    precision : float
        If y_test is set in predict method, calculates precision internally.

    recall : float
        If y_test is set in predict method, calculates recall internally.

    Example
    -------
    >>> import ims
    >>> ds = ims.Dataset.read_mea("IMS_data")
    >>> X_train, X_test, y_train, y_test = ds.train_test_split()
    >>> model = ims.PLS_DA(ds, n_components=5)
    >>> model.fit(X_train, y_train)
    >>> model.predict(X_test, y_test)
    >>> model.plot()
    """
    def __init__(self, dataset, n_components=2, **kwargs):
        self.dataset = dataset
        self.n_components = n_components
        self._sk_pls = PLSRegression(n_components=n_components,
                                     scale=False, **kwargs)

    @staticmethod
    def _create_binary_labels(y):
        """Creates a binary label matrix with one column per group."""
        groups = np.unique(y)
        y_binary = np.zeros((len(y), len(groups)))
        for i, j in enumerate(groups):
            col = [j in label for label in y]
            y_binary[:, i] = col
        return y_binary

    @staticmethod
    def _reverse_binary_labels(y, groups):
        """
        Returns a list of class labels from a binary label matrix
        and a tuple of unique groups.
        """
        y_reversed = []
        for label in y:
            i = np.argmax(label)
            y_reversed.append(groups[i])
        return y_reversed

    def fit(self, X_train, y_train):
        """
        Fits the model with training data.
        Converts the labels into a binary matrix.

        Parameters
        ----------
        X_train : numpy.ndarray of targets (n_samples, n_features)
            Training vectors with features.

        y_train : numpy.ndarray of shape (n_samples,)
            True class labels for training data.

        Returns
        -------
        self
            Fitted model.
        """
        self.groups = np.unique(y_train)
        self.y_train = y_train
        y_binary = self._create_binary_labels(y_train)
        self._sk_pls.fit(X_train, y_binary)
        self.x_scores, self.y_scores = self._sk_pls.transform(X_train, y_binary)
        self.x_weights = self._sk_pls.x_weights_
        self.x_loadings = self._sk_pls.x_loadings_
        self.y_weights = self._sk_pls.y_weights_
        self.y_loadings = self._sk_pls.y_loadings_
        self.coefficients = self._sk_pls.coef_
        self.vip_scores = ims.utils.vip_scores(self.x_weights, self.x_scores, self.y_loadings)
        return self

    def predict(self, X_test, y_test=None):
        """
        Predicts class labels for test data. Converts back from binary
        labels matrix to a list of class names. If y_test is set also calculates
        accuracy, precision and recall and stores them as attributes.

        Parameters
        ----------
        X_test : numpy.ndarray of shape (n_samples, n_features)
            Feature vectors of test dataset.

        y_test : numpy.ndarray of shape (n_samples,), optional
            True labels for test dataset. If set calculates error metrics,
            by default None

        Returns
        -------
        numpy.ndarray of shape (n_samples,)
            Predicted class labels.
        """        
        y_pred_binary = self._sk_pls.predict(X_test)
        y_pred = self._reverse_binary_labels(y_pred_binary, self.groups)

        if y_test is not None:
            self.y_test = y_test
            self.accuracy = accuracy_score(y_test, y_pred)
            self.precision = precision_score(
                y_test, y_pred,
                average="weighted",
                labels=np.unique(y_pred)
                )
            self.recall = recall_score(
                y_test, y_pred,
                average="weighted",
                labels=np.unique(y_pred)
                )

        return np.array(y_pred)
    
    def transform(self, X, y=None):
        """
        Apply the dimensionality reduction.

        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            Feature matrix.
        y : numpy.ndarray of shape (n_samples, n_targtets), optional
            Dependend variables, by default None

        Returns
        -------
        X_scores
        """
        return self._sk_pls.transform(X, y)
    
    def score(self, X_test, y_test, sample_weight=None):
        """
        Calculates accuracy score for predicted data.

        Parameters
        ----------
        X_test : numpy.ndarray of shape (n_samples, n_features)
            Feature vectors of the test data.

        y_test : numpy.ndarray of shape (n_samples,)
            True classification labels.

        sample_weight : array-like of shape (n_samples,), optional
            Sample weights, by default None.

        Returns
        -------
        score : float
            Mean accuracy score.
        """
        y_pred = self.predict(X_test)
        return accuracy_score(y_test, y_pred, sample_weight=sample_weight)

    def plot(self, x_comp=1, y_comp=2, width=9, height=8,
             annotate=False):
        """
        Plots PLS components as scatter plot.

        Parameters
        ----------
        x_comp : int, optional
            Component x axis, by default 1.

        y_comp : int, optional
            Component y axis, by default 2.

        width : int or float, optional
            Width of the plot in inches,
            by default 9.

        height : int or float, optional
            Height of the plot in inches,
            by default 8.

        annotate : bool, optional
            If True adds sample names to markers,
            by default False.

        Returns
        -------
        matplotlib.pyplot.axes
        """
        cols = [f"PLS Component {i}" for i in range(1, self.n_components + 1)]

        df = pd.DataFrame(self.x_scores, columns=cols)
        df["Group"] = self.y_train
        if hasattr(self.dataset, "train_index"):
            df["Sample"] = self.dataset[self.dataset.train_index].samples
        else:
            df["Sample"] = self.dataset.samples

        plt.figure(figsize=(width, height))
        ax = sns.scatterplot(
            x=f"PLS Component {x_comp}",
            y=f"PLS Component {y_comp}",
            data=df,
            hue="Group",
            style="Group",
            )
        plt.legend(frameon=True, fancybox=True, facecolor="white")

        if annotate:
            for _, row in df.iterrows():
                plt.annotate(
                    row["Sample"],
                    xy=(row[f"PLS Component {x_comp}"], row[f"PLS Component {y_comp}"]),
                    xycoords="data"
                )

        return ax
    
    def plot_loadings(self, component=1, color_range=0.02,
                      width=9, height=10):
        """
        Plots PLS x loadings as image with retention and drift
        time coordinates.

        Parameters
        ----------
        component : int, optional
            Component to plot, by default 1.

        color_range : float, optional
            Minimum and Maximum to adjust to different scaling methods,
            by default 0.02.
            
        width : int or float, optional
            Width of the plot in inches,
            by default 9.

        height : int or float, optional
            Height of the plot in inches,
            by default 10.

        Returns
        -------
        matplotlib.pyplot.axes
        """        
        loadings = self.x_loadings[:, component-1].\
            reshape(self.dataset[0].shape)
            
        ret_time = self.dataset[0].ret_time
        drift_time = self.dataset[0].drift_time

        _, ax = plt.subplots(figsize=(width, height))

        plt.imshow(
            loadings,
            cmap="RdBu_r",
            vmin=(-color_range),
            vmax=color_range,
            origin="lower",
            aspect="auto",
            extent=(min(drift_time), max(drift_time),
                    min(ret_time), max(ret_time))
            )

        plt.colorbar(label="PLS-DA loadings")
        plt.title(f"PLS-DA loadings of component {component}")
        plt.xlabel(self.dataset[0]._drift_time_label)
        plt.ylabel("Retention time [s]")
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        return ax

    def plot_coefficients(self, group=0, width=9, height=10):
        """
        Plots PLS coefficients of selected group as image
        with retention and drift time axis.

        Parameters
        ----------
        group : int or str, optional
            Index or name of group, by default 0.

        width : int or float, optional
            Width of the plot in inches,
            by default 9.

        height : int or float, optional
            Height of the plot in inches,
            by default 10.

        Returns
        -------
        matplotlib.pyplot.axes
        """
        if isinstance(group, str):
            group_index = self.groups.index(group)
            group_name = group

        if isinstance(group, int):
            group_index = group
            group_name = self.groups[group]

        coef = self._sk_pls.coef_[:, group_index].\
            reshape(self.dataset[0].values.shape)

        ret_time = self.dataset[0].ret_time
        drift_time = self.dataset[0].drift_time
        
        _, ax = plt.subplots(figsize=(width, height))
        
        plt.imshow(
            coef,
            cmap="RdBu_r",
            norm=CenteredNorm(0),
            origin="lower",
            aspect="auto",
            extent=(min(drift_time), max(drift_time),
                    min(ret_time), max(ret_time))
            )

        plt.colorbar(label="PLS-DA coefficients")
        plt.title(f"PLS-DA coefficients of {group_name}")
        plt.xlabel(self.dataset[0]._drift_time_label)
        plt.ylabel("Retention time [s]")
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        return ax
    
    def plot_vip_scores(self, threshold=None, width=9, height=10):
        """
        Plots VIP scores as image with retention and drift time axis.
        
        Parameters
        ----------
        width : int or float, optional
            Width of the plot in inches,
            by default 9.

        height : int or float, optional
            Height of the plot in inches,
            by default 10.

        Returns
        -------
        matplotlib.pyplot.axes
        """        
        if not hasattr(self, "vip_scores"):
            raise ValueError("Must fit data first.")
        
        if threshold is None:
            vip_matrix = self.vip_scores.reshape(self.dataset[0].values.shape)
        else:
            vips = np.zeros_like(self.vip_scores)
            i = np.where(self.vip_scores > threshold)
            vips[i] = self.vip_scores[i]
            vip_matrix = vips.reshape(self.dataset[0].values.shape)

        ret_time = self.dataset[0].ret_time
        drift_time = self.dataset[0].drift_time

        _, ax = plt.subplots(figsize=(width, height))

        plt.imshow(
            vip_matrix,
            cmap="RdBu_r",
            origin="lower",
            aspect="auto",
            extent=(min(drift_time), max(drift_time),
                    min(ret_time), max(ret_time))
            )

        plt.colorbar(label="VIP scores")
        plt.title(f"PLS-DA VIP scores")
        plt.xlabel(self.dataset[0]._drift_time_label)
        plt.ylabel("Retention time [s]")
        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        return ax



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.ticker import MaxNLocator, AutoMinorLocator
from sklearn.decomposition import PCA


class PCA_Model:
    """
    PCA_Model is a wrapper class around the scikit-learn PCA implementation
    and provides prebuilt plots for GC-IMS datasets.

    See the original scikit-learn documentation for a detailed description:
    https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html

    Parameters
    ----------
    dataset : ims.Dataset
        The dataset is needed for the retention and drift time
        coordinates.

    n_components : int or float, optional
        Number of components to keep. If None all components are kept,
        by default None.

    svd_solver : str, optional
        "auto", "full", "arpack" or "randomised" are valid,
        by default "auto".

    **kwargs: optional
        Additional key word arguments are passed to the scikit-learn PCA.
        See the original documentation for valid parameters.

    Attributes
    ----------
    scores : numpy.ndarray of shape (n_samples, n_features)
        X with dimension reduction applied.

    loadings : numpy.ndarray of shape (n_components, n_features)
        PCA loadings already corrected when a scaling method was
        applied on the dataset.

    explainded_variance : numpy.ndarray of shape (n_components,)
        The amount of variance explained by each component.

    explained_variance_ratio : numpy.nd_array of shape (n_components,)
        Percentage of variance explained by each component.

    singular_values : numpy.ndarray of shape (n_components,)
        The singular values corresponding to each component.

    mean : numpy.ndarray of shape (n_features,)
        Per feature mean estimated from training data.

    Example
    -------
    >>> import ims
    >>> ds = ims.Dataset.read_mea("IMS_data")
    >>> X, _ = ds.get_xy()
    >>> pca = ims.PCA_Model(ds, n_components=20)
    >>> pca.fit(X)
    >>> pca.plot()
    """
    def __init__(self, dataset, n_components=None, svd_solver="auto", **kwargs):
        self.dataset = dataset
        if n_components is None:
            self.n_components = len(self.dataset)
        else:
            self.n_components = n_components
        self.svd_solver = svd_solver
        self._sk_pca = PCA(n_components, svd_solver=svd_solver, **kwargs)

    def fit(self, X_train):
        """
        Fit the PCA model with training data.

        Parameters
        ----------
        X_train : numpy.ndarray of shape (n_samples, n_features)
            The training data.
            
        Returns
        -------
        self
            The fitted model.
        """
        self._sk_pca.fit(X_train)
        self.scores = self._sk_pca.transform(X_train)
        self.explained_variance = self._sk_pca.explained_variance_
        self.explained_variance_ratio = self._sk_pca.explained_variance_ratio_
        self.singular_values = self._sk_pca.singular_values_
        self.mean = self._sk_pca.mean_
        self.loadings = self._sk_pca.components_
        return self

    def plot(self, PC_x=1, PC_y=2, width=9, height=8, annotate=False):
        """
        Scatter plot of selected principal components.

        Parameters
        ----------
        PC_x : int, optional
            PC x axis, by default 1.

        PC_y : int, optional
            PC y axis, by default 2.

        width : int or float, optional
            plot width in inches, by default 8.

        height : int or float, optional
            plot height in inches, by default 7.
            
        annotate : bool, optional
            label data points with sample name,
            by default False.

        Returns
        -------
        matplotlib.pyplot.axes
        """
        expl_var = []
        for i in range(1, self.n_components + 1):
            expl_var.append(round(self.explained_variance_ratio[i-1] * 100, 1))
        
        pc_df = pd.DataFrame(
            data=self.scores,
            columns=[f"PC {x}" for x in range(1, self.n_components + 1)]
        )
        
        if hasattr(self.dataset, "train_index"):
            pc_df["Sample"] = self.dataset[self.dataset.train_index].samples
            pc_df['Label'] = self.dataset[self.dataset.train_index].labels
        else:
            pc_df["Sample"] = self.dataset.samples
            pc_df['Label'] = self.dataset.labels

        _, ax = plt.subplots(figsize=(width, height))
        sns.scatterplot(
            ax=ax,
            x=f"PC {PC_x}",
            y=f"PC {PC_y}",
            data=pc_df,
            hue="Label",
            style="Label",
        )

        plt.legend(frameon=True, fancybox=True, facecolor="white")
        plt.xlabel(f"PC {PC_x} ({expl_var[PC_x-1]} % of variance)")
        plt.ylabel(f"PC {PC_y} ({expl_var[PC_y-1]} % of variance)")

        if annotate:
            for i, point in pc_df.iterrows():
                ax.text(point[f"PC {PC_x}"], point[f"PC {PC_y}"],
                        point["Sample"])

        return ax

    def plot_loadings(self, PC=1, color_range=0.1, width=9, height=10):
        """
        Plots loadings of a principle component with the original retention
        and drift time coordinates.

        Parameters
        ----------
        PC : int, optional
            principal component, by default 1.

        color_range : int, optional
            color_scale ranges from - color_range to + color_range
            centered at 0.

        width : int or float, optional
            plot width in inches, by default 9.

        height : int or float, optional
            plot height in inches, by default 10.

        Returns
        -------
        matplotlib.pyplot.axes
        """        
        # use retention and drift time axis from the first spectrum
        ret_time = self.dataset[0].ret_time
        drift_time = self.dataset[0].drift_time

        loading_pc = self.loadings[PC-1, :].reshape(len(ret_time),
                                                    len(drift_time))
        
        expl_var = []
        for i in range(1, self.n_components + 1):
            expl_var.append(round(self.explained_variance_ratio[i-1] * 100, 1))

        _, ax = plt.subplots(figsize=(width, height))

        plt.imshow(
            loading_pc,
            origin="lower",
            aspect="auto",
            cmap="RdBu_r",
            vmin=(-color_range),
            vmax=color_range,
            extent=(min(drift_time), max(drift_time),
                    min(ret_time), max(ret_time))
            )

        plt.colorbar().set_label("PCA loadings")

        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())
        
        plt.xlabel(self.dataset[0]._drift_time_label)
        plt.ylabel("Retention time [s]")
        plt.title(f"PCA loadings of PC {PC} ({expl_var[PC-1]} % of variance)")
        return ax

    def scree_plot(self, width=9, height=8):
        """
        Plots the explained variance ratio per principal component
        and cumulatively.

        Parameters
        ----------
        width : int or float, optional
            Width of the plot in inches,
            by default 9.

        height : int or float, optional
            Height of the plot in inches,
            by default 8.

        Returns
        -------
        matplotlib.pyplot.axes
        """
        x = [*range(1, self.n_components + 1)]
        y = self.explained_variance_ratio

        _, ax = plt.subplots(figsize=(width, height))
        
        for axis in [ax.xaxis, ax.yaxis]:
            axis.set_major_locator(MaxNLocator(integer=True))
        
        plt.xticks(x)
        plt.xlabel("Principal component")
        plt.ylabel("Explainded variance ratio [%]")
        
        ax.plot(x, np.cumsum(y) * 100, label="cumulative")
        ax.plot(x, y * 100, label="per PC")
        
        plt.legend(frameon=True, fancybox=True, facecolor="white")
        return ax



from ims import Spectrum
import numpy as np
import os
from glob import glob
from copy import deepcopy
from datetime import datetime
import h5py
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d
from sklearn.utils import resample
from sklearn.model_selection import (ShuffleSplit,
    KFold, StratifiedKFold, LeaveOneOut, StratifiedShuffleSplit)


class Dataset:
    """
    Dataset class coordinates many GC-IMS spectra
    (instances of ims.Spectrum class) with labels, file
    and sample names.

    ims.Spectrum methods are applied to all spectra. It also contains
    additional functionality and methods that require multiple spectra
    at a time such as alignments and calculating means. Most operations
    are done inplace for memory efficiency.

    Use one of the read_... methods as constructor.

    Parameters
    ----------
    data : list
        Lists instances of `ims.Spectrum`.

    name : str
        Name of the dataset.

    files : list
        Lists one file name per spectrum. Must be unique.

    samples : list
        Lists sample names. A sample can have multiple files in
        case of repeat determination. Needed to calculate means.

    labels : list or numpy.ndarray
        Classification or regression labels.

    Attributes
    ----------
    preprocessing : list
        Keeps track of applied preprocessing steps.

    weights : numpy.ndarray of shape (n_samples, n_features)
        Stores the weights from scaling when the method is called.
        Needed to correct the loadings in PCA automatically.
        
    train_index : list
        Keeps the indices from train_test_split method.
        Used for plot annotations in PLS_DA and PLSR classes.
        
    test_index : list
        Keeps the indices from train_test_split method.
        Used for plot annotations in PLS_DA and PLSR classes.

    Example    
    -------
    >>> import ims
    >>> ds = ims.Dataset.read_mea("IMS_data")
    >>> print(ds)
    Dataset: IMS_data, 58 Spectra
    """
    def __init__(self, data, name=None, files=None, samples=None, labels=None):
        self.data = data
        self.name = name
        self.files = files
        self.samples = samples
        self.labels = labels
        self.preprocessing = []

    def __repr__(self):
        return f'Dataset: {self.name}, {len(self)} Spectra'

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.data[key]
        
        if isinstance(key, slice):
            return Dataset(
                self.data[key],
                self.name,
                self.files[key],
                self.samples[key],
                self.labels[key]
            )
            
        if isinstance(key, list) or isinstance(key, np.ndarray):
            return Dataset(
                [self.data[i] for i in key],
                self.name,
                [self.files[i] for i in key],
                [self.samples[i] for i in key],
                [self.labels[i] for i in key]
            )
        
    def __delitem__(self, key):
        del self.data[key]
        del self.files[key]
        del self.samples[key]
        del self.labels[key]

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.data)
    
    def copy(self):
        """
        Uses deepcopy from the copy module in the standard library.
        Most operations happen inplace. Use this method if you do not
        want to change the original variable.

        Returns
        -------
        Dataset
            deepcopy of self.

        Example
        -------
        >>> import ims
        >>> ds = ims.Dataset.read_mea("IMS_data")
        >>> new_variable = ds.copy()
        """        
        return deepcopy(self)

    @property
    def sample_indices(self):
        """
        This property returns information about all spectra indices
        for each sample in the dataset.
        Useful to select or remove specific samples or files.

        Returns
        -------
        dict
            Sample names as keys, lists with indices of spectra as values.
        """
        u_samples = np.unique(self.samples)
        indices = []
        for i in u_samples:
            index = np.where(np.array(self.samples) == i)
            indices.append(index)

        indices = [list(i[0]) for i in indices]
        indices = dict(zip(u_samples, indices))
        return indices

    @staticmethod
    def _measurements(path, subfolders):
        """
        Lists paths to every file in folder.
        Optionally generates label and sample names by splitting file paths.
        """
        if subfolders:
            files = []
            samples = []
            labels = []
            paths = [os.path.normpath(i) for i in glob(f'{path}/*/*/*')]
            name = os.path.split(path)[1]
            for filedir in paths:
                file_name = os.path.split(filedir)[1]
                files.append(file_name)
                sample_name = filedir.split(os.sep)[-2]
                samples.append(sample_name)
                label = filedir.split(os.sep)[-3]
                labels.append(label)
        else:
            paths = [os.path.normpath(i) for i in glob(f'{path}/*')]
            name = os.path.split(path)[1]
            files = [os.path.split(i)[1] for i in paths]
            samples = []
            labels = []

        return (paths, name, files, samples, labels)

    @classmethod
    def read_mea(cls, path, subfolders=False):
        """
        Reads all mea files from G.A.S Dortmund instruments in the
        given directory and combines them into a dataset.
        Much faster than reading csv files and therefore preferred.

        If subfolders=True expects the following folder structure
        for each label and sample:

        * Data
            * Group A
                * Sample A
                    * file a
                    * file b
                * Sample B
                    * file a
                    * file b

        Labels can then be auto-generated from directory names.
        Otherwise labels and sample names need to be added from other sources
        for all methods to work.

        Parameters
        ----------
        path : str
            Absolute or relative file path.

        subfolders : bool, optional
            Uses subdirectory names as labels,
            by default False.

        Returns
        -------
        Dataset

        Example
        -------
        >>> import ims
        >>> ds = ims.Dataset.read_mea("IMS_data", subfolders=True)
        >>> print(ds)
        Dataset: IMS_data, 58 Spectra
        """
        paths, name, files, samples, labels = Dataset._measurements(
            path, subfolders
        )
        data = [Spectrum.read_mea(i) for i in paths]
        return cls(data, name, files, samples, labels)

    @classmethod
    def read_zip(cls, path, subfolders=False):
        """
        Reads zipped csv and json files from G.A.S Dortmund mea2zip converting tool.
        Present for backwards compatibility. Reading mea files is much faster and saves
        the manual extra step of converting.
        
        If subfolders=True expects the following folder structure
        for each label and sample:

        * Data
            * Group A
                * Sample A
                    * file a
                    * file b
                * Sample B
                    * file a
                    * file b

        Labels can then be auto-generated from directory names.
        Otherwise labels and sample names need to be added from other sources
        for all methods to work.

        Parameters
        ----------
        path : str
            Absolute or relative file path.

        Returns
        -------
        Dataset

        Example
        -------
        >>> import ims
        >>> ds = ims.Dataset.read_zip("IMS_data", subfolders=True)
        >>> print(ds)
        Dataset: IMS_data, 58 Spectra
        """
        paths, name, files, samples, labels = Dataset._measurements(
            path, subfolders
        )
        data = [Spectrum.read_zip(i) for i in paths]
        return cls(data, name, files, samples, labels)
    
    @classmethod
    def read_csv(cls, path, subfolders=False):
        """
        Reads generic csv files. The first row must be
        the drift time values and the first column must be
        the retention time values. Values inbetween are the
        intensity matrix.
        Uses the time when the file was created as timestamp.
        
        If subfolders=True expects the following folder structure
        for each label and sample:

        * Data
            * Group A
                * Sample A
                    * file a
                    * file b
                * Sample B
                    * file a
                    * file b

        Labels can then be auto-generated from directory names.
        Otherwise labels and sample names need to be added from other sources
        for all methods to work.

        Parameters
        ----------
        path : str
            Absolute or relative file path.

        Returns
        -------
        Dataset

        Example
        -------
        >>> import ims
        >>> ds = ims.Dataset.read_csv("IMS_data", subfolders=True)
        >>> print(ds)
        Dataset: IMS_data, 58 Spectra
        """
        paths, name, files, samples, labels = Dataset._measurements(
            path, subfolders
        )
        data = [Spectrum.read_csv(i) for i in paths]
        return cls(data, name, files, samples, labels)

    @classmethod
    def read_hdf5(cls, path):
        """
        Reads hdf5 files exported by the Dataset.to_hdf5 method.
        Convenient way to store preprocessed spectra.
        Especially useful for larger datasets as preprocessing
        requires more time.
        Preferred to csv because of faster read and write speeds.

        Parameters
        ----------
        path : str
            Absolute or relative file path.

        Returns
        -------
        Dataset

        Example
        -------
        >>> import ims
        >>> sample = ims.Dataset.read_mea("IMS_data")
        >>> sample.to_hdf5("IMS_data_hdf5")
        >>> sample = ims.Dataset.read_hdf5("IMS_data_hdf5")
        """
        with h5py.File(path, "r") as f:
            labels = [i.decode() for i in f["dataset"]["labels"]]
            samples = [i.decode() for i in f["dataset"]["samples"]]
            files = [i.decode() for i in f["dataset"]["files"]]
            preprocessing = [i.decode() for i in f["dataset"]["preprocessing"]]

            data = []
            for key in f.keys():
                if key == "dataset":
                    continue
                values = np.array(f[key]["values"])
                ret_time = np.array(f[key]["ret_time"])
                drift_time = np.array(f[key]["drift_time"])
                name = str(f[key].attrs["name"])
                time = datetime.strptime(f[key].attrs["time"],
                                        "%Y-%m-%dT%H:%M:%S")
                drift_time_label = str(f[key].attrs["drift_time_label"])
                spectrum = Spectrum(name, values, ret_time, drift_time, time)
                spectrum._drift_time_label = drift_time_label
                data.append(spectrum)

            name = os.path.split("Test.hdf5")[1]
            name = name.split('.')[0]

        dataset = cls(data, name, files, samples, labels)
        dataset.preprocessing = preprocessing
        return dataset

    def to_hdf5(self, name=None, path=None):
        """
        Exports the dataset as hdf5 file.
        It contains one group per spectrum and one with labels etc.
        Use ims.Dataset.read_hdf5 to read the file and construct a dataset.

        Parameters
        ----------
        name : str, optional
            Name of the hdf5 file. File extension is not needed.
            If not set, uses the dataset name attribute,
            by default None.
            
        path : str, otional
            Path to save the file. If not set uses the current working
            directory, by default None.

        Example
        -------
        >>> import ims
        >>> ds = ims.Dataset.read_mea("IMS_data")
        >>> ds.to_hdf5()
        >>> ds = ims.Dataset.read_hdf5("IMS_data.hdf5")
        """
        if name is None:
            name = self.name
 
        if path is None:
            path = os.getcwd()

        with h5py.File(f"{path}/{name}.hdf5", "w-") as f:
            data = f.create_group("dataset")
            data.create_dataset("labels", data=self.labels)
            data.create_dataset("samples", data=self.samples)
            data.create_dataset("files", data=self.files)
            data.create_dataset("preprocessing", data=self.preprocessing)

            for sample in self:
                grp = f.create_group(sample.name)
                grp.attrs["name"] = sample.name
                grp.create_dataset("values", data=sample.values)
                grp.create_dataset("ret_time", data=sample.ret_time)
                grp.create_dataset("drift_time", data=sample.drift_time)
                grp.attrs["time"] = datetime.strftime(sample.time,
                                                    "%Y-%m-%dT%H:%M:%S")
                grp.attrs["drift_time_label"] = sample._drift_time_label

    def select(self, label=None, sample=None):
        """
        Selects all spectra of specified label or sample.
        Must set at least one of the parameters.

        Parameters
        ----------
        label : str, optional
            Label name to keep, by default None

        sample : str, optional
            Sample name to keep, by default None

        Returns
        -------
        Dataset
            Contains only matching spectra.

        Example
        -------
        >>> import ims
        >>> ds = ims.Dataset.read_mea("IMS_data")
        >>> group_a = ds.select(label="GroupA") 
        """
        if label is None and sample is None:
            raise ValueError("Must give either label or sample value.")
        
        if label is not None:
            name = label
            indices = []
            for i, j in enumerate(self.labels):
                if j == label:
                    indices.append(i)

        if sample is not None:
            name = sample
            indices = []
            for i, j in enumerate(self.samples):
                if j == sample:
                    indices.append(i)

        result = []
        files = []
        labels = []
        samples = []
        for i in indices:
            result.append(self.data[i])
            files.append(self.files[i])
            labels.append(self.labels[i])
            samples.append(self.samples[i])

        return Dataset(
            data=result,
            name=name,
            files=files,
            samples=samples,
            labels=labels,
        )
        
    def drop(self, label=None, sample=None):
        """
        Removes all spectra of specified label or sample from dataset.
        Must set at least one of the parameters.

        Parameters
        ----------
        label : str, optional
            Label name to keep, by default None

        sample : str, optional
            Sample name to keep, by default None

        Returns
        -------
        Dataset
            Contains only matching spectra.

        Example
        -------
        >>> import ims
        >>> ds = ims.Dataset.read_mea("IMS_data")
        >>> ds = ds.drop(label="GroupA")
        """
        if label is None and sample is None:
            raise ValueError("Must give either label or sample value.")

        if label is not None:
            indices = []
            for i, j in enumerate(self.labels):
                if j != label:
                    indices.append(i)

        if sample is not None:
            indices = []
            for i, j in enumerate(self.samples):
                if j != sample:
                    indices.append(i)

        result = []
        files = []
        labels = []
        samples = []
        for i in indices:
            result.append(self.data[i])
            files.append(self.files[i])
            labels.append(self.labels[i])
            samples.append(self.samples[i])

        return Dataset(
            data=result,
            name=self.name,
            files=files,
            samples=samples,
            labels=labels,
        )
        
    def add(self, spectrum, sample, label):
        """
        Adds a ims.Spectrum to the dataset.
        Sample name and label must be provided because they are
        not stored in the ims.Spectrum class.

        Parameters
        ----------
        spectrum : ims.Spectrum
            GC-IMS spectrum to add to the dataset.

        sample : str
            The sample name is added to the sample attribute.
            Necessary because sample names are not stored in ims.Spectrum class.

        label : various
            Classification or regression label is added to the label attribute.
            Necessary because labels are not stored in ims.Spectrum class.

        Returns
        -------
        Dataset
            With Spectrum added.

        Example
        -------
        >>> import ims
        >>> ds = ims.Dataset.read_mea("IMS_data")
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> ds.add(sample, "sample_name", "class_label")
        """
        self.data.append(spectrum)
        self.files.append(spectrum.name)
        self.samples.append(sample)
        self.labels.append(label)
        return self

    def groupby(self, key="label"):
        """
        Groups dataset by label or sample.

        Parameters
        ----------
        key : str, optional
            "label" or "sample" are valid keys, by default "label"

        Returns
        -------
        list
            List of one ims.Dataset instance per group or sample.
        """
        if key != "label" and key != "sample":
            raise ValueError('Only "label" or "sample" are valid keys!')
            
        result = []
        if key == "label":
            for group in np.unique(self.labels):
                result.append(self.select(label=group))
            return result

        if key == "sample":
            for sample in np.unique(self.samples):
                result.append(self.select(sample=sample))
            return result
        
    def plot(self, index=0, **kwargs):
        """
        Plots the spectrum of selected index and adds the label to the title.

        Parameters
        ----------
        index : int, optional
            Index of spectrum to plot, by default 0

        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
        """
        ax = self[index].plot(**kwargs)
        plt.title(f"{self[index].name}; {self.labels[index]}")
        return ax

    def train_test_split(self, test_size=0.2, stratify=False, random_state=None):
        """
        Splits the dataset in train and test sets.

        Parameters
        ----------
        test_size : float, optional
            Proportion of the dataset to be used for validation.
            Should be between 0.0 and 1.0,
            by default 0.2
            
        stratify : bool, optional
            Wheter to stratify output or not.
            Preserves the percentage of samples from each class in each split,
            by default False.

        random_state : int, optional
            Controls the randomness. Pass an int for reproducible output,
            by default 1

        Returns
        -------
        tuple of numpy.ndarray
            X_train, X_test, y_train, y_test
            
        Example
        -------
        >>> import ims
        >>> ds = ims.Dataset.read_mea("IMS_Data")
        >>> X_train, X_test, y_train, y_test = ds.train_test_split()
        """
        
        if stratify:
            s = StratifiedShuffleSplit(
                n_splits=1,
                test_size=test_size,
                random_state=random_state
                )
            train, test = next(s.split(self.data, y=self.labels))
        else:
            s = ShuffleSplit(
                n_splits=1,
                test_size=test_size,
                random_state=random_state
                )
            train, test = next(s.split(self.data))

        # set attributes for annotations of plots with sample names
        self.train_index = train
        self.test_index = test

        X_train, y_train = self[train].get_xy()
        X_test, y_test = self[test].get_xy()
        return X_train, X_test, y_train, y_test
    
    def kfold_split(self, n_splits=5, shuffle=True,
                    random_state=None, stratify=False):
        """
        K-Folds cross-validator (sklearn.model_selection.KFold).
        Splits the dataset into k consecutive folds and provides
        train and test data.

        If stratify is True uses StratifiedKfold instead.

        Parameters
        ----------
        n_splits : int, optional
            Number of folds. Must be at least 2,
            by default 5.

        shuffle : bool, optional
            Whether to shuffle the data before splitting,
            by default True.

        random_state : int, optional
            When shuffle is True random_state affects the order of the indices.
            Pass an int for reproducible splits,
            by default None.

        stratify : bool, optional
            Wheter to stratify output or not.
            Preserves the percentage of samples from each class in each split,
            by default False.

        Yields
        ------
        tuple
            (X_train, X_test, y_train, y_test) per iteration

        Example
        -------
        >>> import ims
        >>> from sklearn.metrics import accuracy_score
        >>> ds = ims.Dataset.read_mea("IMS_data")
        >>> model = ims.PLS_DA(ds)
        >>> accuracy = []
        >>> for X_train, X_test, y_train, y_test in ds.kfold_split():
        >>>     model.fit(X_train, y_train)
        >>>     y_pred = model.predict(X_test)
        >>>     accuracy.append(accuracy_score(y_test, y_pred))
        """
        if stratify:
            kf = StratifiedKFold(n_splits=n_splits, shuffle=shuffle,
                                 random_state=random_state)
        else:
            kf = KFold(n_splits, shuffle=shuffle, random_state=random_state)

        for train_index, test_index in kf.split(self, self.labels):
            train_data = self[train_index]
            test_data = self[test_index]
            X_train, y_train = train_data.get_xy()
            X_test, y_test = test_data.get_xy()
            yield X_train, X_test, y_train, y_test
            
    def shuffle_split(self, n_splits=5, test_size=0.2, random_state=None):
        """
        Shuffled splits for montecarlo cross-validation. Randomly selects
        a fraction of the dataset, without replacements, per split
        (sklearn.model_selection.ShuffleSplit).

        Parameters
        ----------
        n_splits : int, optional
            Number of re-shuffling and splitting iterations,
            by default 5.

        test_size : float, optional
            Proportion of the dataset to include in the test split,
            by default 0.2.

        random_state : int, optional
            Controls randomness. Pass an int for reproducible output,
            by default None.

        Yields
        -------
        tuple
            (X_train, X_test, y_train, y_test) per iteration
            
        Example
        -------
        >>> import ims
        >>> from sklearn.metrics import accuracy_score
        >>> ds = ims.Dataset.read_mea("IMS_data")
        >>> model = ims.PLS_DA(ds)
        >>> accuracy = []
        >>> for X_train, X_test, y_train, y_test in ds.shuffle_split():
        >>>     model.fit(X_train, y_train)
        >>>     y_pred = model.predict(X_test)
        >>>     accuracy.append(accuracy_score(y_test, y_pred))
        """        
        rs = ShuffleSplit(
            n_splits=n_splits,
            test_size=test_size,
            random_state=random_state
            )
        for train_index, test_index in rs.split(self, self.labels):
            train_data = self[train_index]
            test_data = self[test_index]
            X_train, y_train = train_data.get_xy()
            X_test, y_test = test_data.get_xy()
            yield X_train, X_test, y_train, y_test

    def bootstrap(self, n_bootstraps=5, n_samples=None, random_state=None):
        """
        Iteratively resamples dataset with replacement. Samples can
        be included multiple times or not at all in the training data.
        Uses all samples that are not present in the training data as test data.

        Parameters
        ----------
        n_bootstraps : int, optional
            Number of iterations, by default 5.
            
        n_samples : int, optional
            Number of samples to draw per iteration. Is set to
            the lenghth of the dataset if None,
            by default None.

        random_state : int, optional
            Controls randomness, pass an int for reproducible output,
            by default None.

        Yields
        -------
        tuple
            (X_train, X_test, y_train, y_test) per iteration
            
        Example
        -------
        >>> import ims
        >>> from sklearn.metrics import accuracy_score
        >>> ds = ims.Dataset.read_mea("IMS_data")
        >>> model = ims.PLS_DA(ds)
        >>> accuracy = []
        >>> for X_train, X_test, y_train, y_test in ds.bootstrap():
        >>>     model.fit(X_train, y_train)
        >>>     y_pred = model.predict(X_test)
        >>>     accuracy.append(accuracy_score(y_test, y_pred))
        """
        for _ in range(n_bootstraps):
            train_data, train_labels = resample(
                self.data,
                self.labels,
                n_samples=n_samples,
                random_state=random_state
                )

            test_data = []
            test_labels = []
            for i, j in enumerate(self.data):
                if j not in train_data:
                    test_data.append(j)
                    test_labels.append(self.labels[i])
            X_train, y_train = Dataset(train_data, labels=train_labels).get_xy()
            X_test, y_test = Dataset(test_data, labels=test_labels).get_xy()
            yield X_train, X_test, y_train, y_test

    def leave_one_out(self):
        """
        Leave-One-Out cross-validator.
        Provides train test splits and uses each sample once as test set
        while the remaining data is used for training.

        Yields
        -------
        tuple
            X_train, X_test, y_train, y_test

        Example
        -------
        >>> import ims
        >>> from sklearn.metrics import accuracy_score
        >>> ds = ims.Dataset.read_mea("IMS_data")
        >>> model = ims.PLS_DA(ds)
        >>> accuracy = []
        >>> for X_train, X_test, y_train, y_test in ds.leave_one_out():
        >>>     model.fit(X_train, y_train)
        >>>     y_pred = model.predict(X_test, y_test)
        >>>     accuracy.append(accuracy_score(y_test, y_pred))
        """
        loo = LeaveOneOut()
        for train_index, test_index in loo.split(self):
            train_data = self[train_index]
            test_data = self[test_index]
            X_train, y_train = train_data.get_xy()
            X_test, y_test = test_data.get_xy()
            yield X_train, X_test, y_train, y_test

    def mean(self):
        """
        Calculates means for each sample, in case of repeat determinations.
        Automatically determines which file belongs to which sample.
        Sample names are used for mean spectra and file names are no longer needed.

        Returns
        -------
        Dataset
            With mean spectra.
        """
        indices = self.sample_indices
        u_samples = np.unique(self.samples)

        labels = []
        grouped_data = []
        for i in u_samples:
            label = self.labels[indices[i][0]]
            labels.append(label)

            data = []
            index = indices[i]
            for j in index:
                data.append(self.data[j])
            grouped_data.append(data)

        means = []
        for i in grouped_data:
            means.append(sum(i) / len(i))
            
        for i, j in zip(means, u_samples):
            i.name = j

        self.data = means
        self.samples = list(u_samples)
        self.labels = labels
        self.preprocessing.append('mean()')
        return self

    def tophat(self, size=15):
        """
        Applies white tophat filter on data matrix as a baseline correction.
        Size parameter is the diameter of the circular structuring element.
        (Slow with large size values.)

        Parameters
        ----------
        size : int, optional
            Size of structuring element, by default 15.

        Returns
        -------
        ims.Dataset
        """
        self.data = [Spectrum.tophat(i, size) for i in self.data]
        self.preprocessing.append('tophat')
        return self

    def sub_first_rows(self, n=1):
        """
        Subtracts first row from every row in spectrum.
        Effective and simple baseline correction
        if RIP tailing is a concern but can hide small peaks.

        Returns
        -------
        Dataset
        """
        self.data = [Spectrum.sub_first_rows(i, n) for i in self.data]
        self.preprocessing.append('sub_first_row')
        return self

    def interp_riprel(self):
        """
        Interpolates all spectra to common RIP relative drift time coordinate.
        Alignment along drift time coordinate.

        Returns
        -------
        Dataset
            With RIP relative spectra.
        """
        dt_riprel = []
        interp_fn = []
        for i in self.data:
            dt = i.drift_time
            rip = np.median(np.argmax(i.values, axis=1)).astype('int32')
            rip_ms = np.mean(dt[rip])
            riprel = dt / rip_ms
            f = interp1d(riprel, i.values, axis=1, kind='cubic')
            dt_riprel.append(riprel)
            interp_fn.append(f)

        start = max([i[0] for i in dt_riprel])
        end = min([i[-1] for i in dt_riprel])
        interv = np.median([(i[-1]-i[0]) / len(i) for i in dt_riprel])
        new_dt = np.arange(start, end, interv)

        for i, f in zip(self.data, interp_fn):
            i.values[:, :len(new_dt)]
            i.values = f(new_dt)
            i.drift_time = new_dt
            i._drift_time_label = "Drift time RIP relative"
        
        self.preprocessing.append("interp_riprel()")
        return self

    def rip_scaling(self):
        """
        Scales values relative to global maximum.
        Can be useful to directly compare spectra from
        instruments with different sensitivity. 

        Returns
        -------
        Dataset
            With scaled values.
        """
        self.data = [Spectrum.rip_scaling(i) for i in self.data]
        self.preprocessing.append('rip_scaling')
        return self
    
    def resample(self, n=2):
        """
        Resamples each spectrum by calculating means of every n rows.
        If the length of the retention time is not divisible by n
        it and the data matrix get cropped by the remainder at the long end.

        Parameters
        ----------
        n : int, optional
            Number of rows to mean,
            by default 2.

        Returns
        -------
        Dataset
            Resampled values.
            
        Example
        -------
        >>> import ims
        >>> ds = ims.Dataset.read_mea("IMS_Data")
        >>> print(ds[0].shape)
        (4082, 3150)
        >>> ds.resample(2)
        >>> print(ds[0].shape)
        (2041, 3150)
        """   
        self.data = [Spectrum.resample(i, n) for i in self.data]
        self.preprocessing.append(f'resample({n})')
        return self
    
    def binning(self, n=2):
        """
        Downsamples each spectrum by binning the array with factor n.
        Similar to Spectrum.resampling but works on both dimensions
        simultaneously.
        If the dimensions are not divisible by the binning factor
        shortens it by the remainder at the long end.
        Very effective data reduction because a factor n=2 already 
        reduces the number of features to a quarter.

        Parameters
        ----------
        n : int, optional
            Binning factor, by default 2.

        Returns
        -------
        Dataset
            Downsampled data matrix.

        Example
        -------
        >>> import ims
        >>> ds = ims.Dataset.read_mea("IMS_Data")
        >>> print(ds[0].shape)
        (4082, 3150)
        >>> ds.binning(2)
        >>> print(ds[0].shape)
        (2041, 1575)
        """
        self.data = [Spectrum.binning(i, n) for i in self.data]
        self.preprocessing.append(f'binning({n})')
        return self

    def cut_dt(self, start, stop=None):
        """
        Cuts data along drift time coordinate.
        Range in between start and stop is kept.
        If stop is not given uses the end of the array instead.
        Combination with RIP relative drift time
        values makes it easier to cut the RIP away and focus
        on the peak area.

        Parameters
        ----------
        start : int or float
            Start value on drift time coordinate.
        
        stop : int or float, optional
            Stop value on drift time coordinate.
            If None uses the end of the array,
            by default None.

        Returns
        -------
        Dataset
            New drift time range.

        Example
        -------
        >>> import ims
        >>> ds = ims.Dataset.read_mea("IMS_data")
        >>> print(ds[0].shape)
        (4082, 3150)
        >>> ds.interp_riprel().cut_dt(1.05, 2)
        >>> print(ds[0].shape)
        (4082, 1005)
        """
        self.data = [Spectrum.cut_dt(i, start, stop) for i in self.data]
        self.preprocessing.append(f'cut_dt({start}, {stop})')
        return self

    def cut_rt(self, start, stop=None):
        """
        Cuts data along retention time coordinate.
        Range in between start and stop is kept.
        If stop is not given uses the end of the array instead.

        Parameters
        ----------
        start : int or float
            Start value on retention time coordinate.

        stop : int or float, optional
            Stop value on retention time coordinate.
            If None uses the end of the array,
            by default None.

        Returns
        -------
        Dataset
            New retention time range.

        Example
        -------
        >>> import ims
        >>> ds = ims.Dataset.read_mea("IMS_data")
        >>> print(ds[0].shape)
        (4082, 3150)
        >>> sample.cut_rt(80, 500)
        >>> print(ds[0].shape)
        (2857, 3150)
        """
        self.data = [Spectrum.cut_rt(i, start, stop) for i in self.data]
        self.preprocessing.append(f'cut_rt({start}, {stop})')
        return self

    def export_plots(self, folder_name=None, file_format='jpg', **kwargs):
        """
        Saves a figure per spectrum as image file. See the docs for
        matplotlib savefig function for supported file formats and kwargs
        (https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.savefig.html).

        Creates a new folder for the plots in the current working directory.

        Parameters
        ----------
        folder_name : str, optional
            New directory to save the images to.

        file_format : str, optional
            See matplotlib savefig docs for information
            about supported formats,
            by default 'jpeg'
            
        Example
        -------
        >>> import ims
        >>> ds = ims.Dataset.read_mea("IMS_data")
        >>> ds.export_plots("IMS_data_plots")
        """
        if folder_name is None:
            folder_name = self.name.join("_plots")

        os.mkdir(folder_name)

        for i in self.data:
            i.export_plot(path=folder_name, file_format=file_format,
                            **kwargs)

    def get_xy(self, flatten=True):
        """
        Returns features (X) and labels (y) as numpy arrays.

        Parameters
        ----------
        flatten : bool, optional
            Flattens 3D datasets to 2D, by default True

        Returns
        -------
        tuple
            (X, y)

        Example
        -------
        >>> import ims
        >>> ds = ims.Dataset.read_mea("IMS_data")
        >>> X, y = ds.get_xy()
        """
        X = [i.values for i in self.data]
        X = np.stack(X)
        y = np.array(self.labels)

        if flatten:
            a, b, c = X.shape
            X = X.reshape(a, b*c)

        return (X, y)

    def scaling(self, method="pareto", mean_centering=True):
        """
        Scales and mean centeres features according to selected method.

        Parameters
        ----------
        method : str, optional
            "pareto", "auto" or "var" are valid,
            by default "pareto".
            
        mean_centering : bool, optional
            If true center the data before scaling,
            by default True.

        Returns
        -------
        Dataset

        Raises
        ------
        ValueError
            If scaling method is not supported.
        """
        X = [i.values for i in self.data]
        X = np.stack(X)
        a, b, c = X.shape
        X = X.reshape(a, b*c)

        if method == "auto":
            weights = 1 / np.std(X, 0)
        elif method == "pareto":
            weights = 1 / np.sqrt(np.std(X, 0))
        elif method == "var":
            weights = 1 / np.var(X, 0)
        else:
            raise ValueError(f'{method} is not a supported method!')

        weights = np.nan_to_num(weights, posinf=0, neginf=0)

        if mean_centering:
            X = (X - np.mean(X, 0)) * weights
        else:
            X = X * weights

        for i, j in enumerate(self.data):
            j.values = X[i, :].reshape(b, c)

        self.weights = weights
        self.preprocessing.append(f"scaling({method})")
        return self



import numpy as np


def vip_scores(W, T, Q):
    """
    Calculates variable importance in projection (VIP) scores
    from PLS X weights, X scores, and y loadings.

    Parameters
    ----------
    W : numpy.ndarray of shape (n_features, n_components)
        X weights

    T : numpy.ndarray of shape (n_samples, n_components)
        X scores

    Q : numpy.ndarray of shape (n_targes, n_components)
        y loadings

    Returns
    -------
    numpy.ndarray of shape (n_features,)
        VIP scores.

    References
    ----------
    Farrés, M., Platikanov, S., Tsakovski, S., and Tauler, R. (2015)
    Comparison of the variable importance in projection (VIP)
    and of the selectivity ratio (SR) methods for variable selection and interpretation.
    J. Chemometrics, 29: 528– 536. doi: 10.1002/cem.2736
    
    https://analyticalsciencejournals.onlinelibrary.wiley.com/doi/full/10.1002/cem.2736
    """
    W0 = W / np.sqrt(np.sum(W**2, 0))
    p, _ = W.shape
    sum_sq = np.sum(T**2, 0) * np.sum(Q**2, 0)
    vips = np.sqrt(p * np.sum(sum_sq * (W0**2), 1) / np.sum(sum_sq))
    return vips


def selectivity_ratio(X, B):
    """
    Calculates the selectivity ratio (SR) from PLS decomposition.

    Parameters
    ----------
    X : numpy.ndarray of size (n_samples, n_features)
        Feature matrix.

    B : numpy.ndarray of size (n_features, n_targets)
        PLS coefficients.

    Returns
    -------
    numpy.ndarray of size (n_features,)
        Selectivity ratio.

    References
    ----------
    Farrés, M., Platikanov, S., Tsakovski, S., and Tauler, R. (2015)
    Comparison of the variable importance in projection (VIP)
    and of the selectivity ratio (SR) methods for variable selection and interpretation.
    J. Chemometrics, 29: 528– 536. doi: 10.1002/cem.2736
    
    https://analyticalsciencejournals.onlinelibrary.wiley.com/doi/full/10.1002/cem.2736
    """
    t_tp = X @ B / np.linalg.norm(B)
    p_tp = X.T @ t_tp / (t_tp.T @ t_tp)

    exp_var = (t_tp @ p_tp.T)**2
    exp_res = (X - exp_var)**2

    ss_expl = np.sum(exp_var, axis=0)
    ss_res = np.sum(exp_res, axis=0)
    return ss_expl / ss_res



"""
Module for chemometrics with GC-IMS data
======================================

Documentation is available via docstrings in classes and functions.

Provides:
---------
* Fast IO including a mea file reader
(standard file format for G.A.S Dortmund instruments).

* Preprocessing steps and utilities (alignment, resampling, plotting etc.).

* Scripted statistical workflows with prebuilt plots
for common algorithms.
"""
__version__ = "0.1.2"
__author__ = "Competency Center for Chemometrics Mannheim"
__credits__ = "Competency Center for Chemometrics Mannheim"

from ims.gcims import Spectrum
from ims.dataset import Dataset
from ims.pca import PCA_Model
from ims.plsr import PLSR
from ims.plsda import PLS_DA
from ims.hca import HCA
import ims.utils


import os
import re
import json
import h5py
from array import array
from copy import deepcopy
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator
from datetime import datetime
from time import ctime
from skimage.morphology import white_tophat, disk
from zipfile import ZipFile


class Spectrum:
    """
    Represents one GCIMS-Spectrum with the data matrix,
    retention and drift time coordinates.
    Sample or file name and timestamp are included unique identifiers.
    
    This class contains all methods that can be applied on a per spectrum basis,
    like I/O, plotting and some preprocessing tools. Methods that return a Spectrum change the instance inplace. Use the copy method.

    Use one of the read_... methods as constructor.

    Parameters
    ----------
    name : str
        File or sample name as a unique identifier.
        Reader methods set this attribute to the file name without extension.

    values : numpy.array
        Intensity matrix.
        
    ret_time : numpy.array
        Retention time coordinate.

    drift_time : numpy.array
        Drift time coordinate.

    time : datetime object
        Timestamp when the spectrum was recorded.
        
    Example
    -------
    >>> import ims
    >>> sample = ims.Spectrum.read_mea("sample.mea")
    >>> sample.plot()
    """
    def __init__(self, name, values, ret_time, drift_time, time):
        self.name = name
        self.values = values
        self.ret_time = ret_time
        self.drift_time = drift_time
        self.time = time
        self._drift_time_label = 'Drift time [ms]'
        
    def __repr__(self):
        return f"GC-IMS Spectrum: {self.name}"
    
    # add, radd and truediv are implemented to calculate means easily
    def __add__(self, other):
        values = self.values + other.values
        ret_time = self.ret_time + other.ret_time
        drift_time = self.drift_time + other.drift_time
        x =  Spectrum(self.name, values, ret_time, drift_time,
                      self.time)
        x._drift_time_label = self._drift_time_label
        return x

    def __radd__(self, other):
        if other == 0:
            return self
        else:
            raise NotImplementedError()

    def __truediv__(self, other):
        if isinstance(other, int):
            values = self.values / other
            ret_time = self.ret_time / other
            drift_time = self.drift_time / other
            x = Spectrum(self.name, values, ret_time,
                         drift_time, self.time)
            x._drift_time_label = self._drift_time_label
            return x
        else:
            raise NotImplementedError()

    @property
    def shape(self):
        """
        Shape property of the data matrix.
        Equivalent to `ims.Spectrum.values.shape`.
        """
        return self.values.shape

    def copy(self):
        """
        Uses deepcopy from the copy module in the standard library.
        Most operations happen inplace. Use this method if you do not
        want to change the original variable.

        Returns
        -------
        Spectrum
            deepcopy of self.

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> new_variable = sample.copy()
        """
        return deepcopy(self)

    @classmethod
    def read_zip(cls, path):
        """
        Reads zipped csv and json files from G.A.S Dortmund mea2zip converting tool.
        Present for backwards compatibility.
        Reading mea files is much faster and saves the manual extra step of converting.

        Parameters
        ----------
        path : str
            Absolute or relative file path.

        Returns
        -------
        Spectrum

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_zip("sample.mea")
        >>> print(sample)
        GC-IMS Spectrum: sample
        """
        with ZipFile(path) as myzip:
            with myzip.open('csv_data.csv', 'r') as mycsv:
                values = pd.read_csv(mycsv, header=None)
            with myzip.open('meta_attributes.json', 'r') as myjson:
                meta_attr = json.load(myjson)

        values = np.array(values)
        values = np.delete(values, -1, axis=1)

        ret_time = np.arange(meta_attr['Chunks count'])\
            * (meta_attr['Chunk averages'] + 1)\
            * meta_attr['Chunk trigger repetition [ms]']\
            / 1000
        drift_time = np.arange(meta_attr['Chunk sample count'])\
            / meta_attr["Chunk sample rate [kHz]"]

        path = os.path.normpath(path)
        name = os.path.split(path)[1]
        time = datetime.strptime(meta_attr["Timestamp"],
                                 "%Y-%m-%dT%H:%M:%S")

        return cls(name, values, ret_time, drift_time, time)

    @classmethod
    def read_mea(cls, path):
        """
        Reads mea files from G.A.S Dortmund instruments.
        Alternative constructor for ims.Spectrum class.
        Much faster than reading csv files and therefore preferred.

        Parameters
        ----------
        path : str
            Absolute or relative file path.

        Returns
        -------
        Spectrum

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> print(sample)
        GC-IMS Spectrum: sample
        """
        path = os.path.normpath(path)
        name = os.path.split(path)[1]
        name = name.split('.')[0]

        with open(path, "rb") as f:
            content = f.read()
            i = content.index(0)
            meta_attr = content[:i-1]
            meta_attr = meta_attr.decode("windows-1252")
            data = content[i+1:]
            data = array('h', data)

        meta_attr = meta_attr.split("\n")

        key_re = re.compile("^.*?(?==)")
        value_re = re.compile("(?<==)(.*?)(?=\[|$)")
        # unit_re = re.compile("\[(.*?)\]")

        for i in meta_attr:
            key = key_re.search(i).group(0).strip()
            value = value_re.search(i).group(0).strip()
            if "Chunks count" in key:
                chunks_count = int(value)
            elif "Chunk averages" in key:
                chunk_averages = int(value)
            elif "Chunk sample count" in key:
                chunk_sample_count = int(value)
            elif "Chunk sample rate" in key:
                chunk_sample_rate = int(value)
            elif "Chunk trigger repetition" in key:
                chunk_trigger_repetition = int(value)
            elif "Timestamp" in key:
                timestamp = datetime.strptime(value, '"%Y-%m-%dT%H:%M:%S"')

        data = np.array(data)
        data = data.reshape(chunks_count, chunk_sample_count)

        ret_time = np.arange(chunks_count) * (chunk_averages + 1)\
            * chunk_trigger_repetition / 1000

        drift_time = np.arange(chunk_sample_count) / chunk_sample_rate

        return cls(name, data, ret_time, drift_time, timestamp)

    @classmethod
    def read_csv(cls, path):
        """
        Reads generic csv files. The first row must be
        the drift time values and the first column must be
        the retention time values. Values inbetween are the
        intensity matrix.
        Uses the time when the file was created as timestamp.

        Parameters
        ----------
        path : str
            Absolute or relative file path.

        Returns
        -------
        Spectrum
        
        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_csv("sample.csv")
        >>> print(sample)
        GC-IMS Spectrum: sample
        """
        name = os.path.split(path)[1]
        name = name.split('.')[0]
        df = pd.read_csv(path)
        values = df.values
        ret_time = df.index
        drift_time= df.columns
        timestamp = os.path.getctime(path)
        timestamp = ctime(timestamp)
        timestamp = datetime.strptime(timestamp,
                                      "%a %b  %d %H:%M:%S %Y")
        return cls(name, values, ret_time, drift_time, timestamp)

    @classmethod
    def read_hdf5(cls, path):
        """
        Reads hdf5 files exported by the to_hdf5 method.
        Convenient way to store preprocessed spectra.
        Especially useful for larger datasets as preprocessing
        requires more time.
        Preferred to csv because of very fast read and write speeds.

        Parameters
        ----------
        path : str
            Absolute or relative file path.

        Returns
        -------
        Spectrum

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> sample.to_hdf5()
        >>> sample = ims.Spectrum.read_hdf5("sample.hdf5")
        """     
        with h5py.File(path, 'r') as f:
            values = np.array(f['values'])
            ret_time = np.array(f['ret_time'])
            drift_time = np.array(f['drift_time'])
            name = str(f.attrs['name'])
            time = datetime.strptime(f.attrs['time'], "%Y-%m-%dT%H:%M:%S")
            drift_time_label = str(f.attrs["drift_time_label"])

        spectrum = cls(name, values, ret_time, drift_time, time)
        spectrum._drift_time_label = drift_time_label
        return spectrum

    def to_hdf5(self, path=None):
        """
        Exports spectrum as hdf5 file.
        Useful to save preprocessed spectra, especially for larger datasets.
        Preferred to csv format because of very fast read and write speeds.

        Parameters
        ----------
        path : str, optional
            Directory to export files to,
            by default the current working directory.

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> sample.to_hdf5()
        >>> sample = ims.Spectrum.read_hdf5("sample.hdf5")
        """
        if path is None:
            path = os.getcwd()
        
        with h5py.File(f'{path}/{self.name}.hdf5', 'w-') as f:
            f.create_dataset('values', data=self.values)
            f.create_dataset('ret_time', data=self.ret_time)
            f.create_dataset('drift_time', data=self.drift_time)
            f.attrs['name'] = self.name
            f.attrs['time'] = datetime.strftime(self.time,
                                                "%Y-%m-%dT%H:%M:%S")
            f.attrs['drift_time_label'] = self._drift_time_label

    def tophat(self, size=15):
        """
        Applies white tophat filter on data matrix as a baseline correction.
        Size parameter is the diameter of the circular structuring element.
        (Slow with large size values.)

        Parameters
        ----------
        size : int, optional
            Size of structuring element, by default 15

        Returns
        -------
        Spectrum
        """
        self.values = white_tophat(self.values, disk(size))
        return self

    def sub_first_rows(self, n=1):
        """
        Subtracts first n rows from every row in spectrum.
        Effective and simple baseline correction
        if RIP tailing is a concern but can hide small peaks.

        Returns
        -------
        Spectrum
        """
        fl = self.values[0:n-1, :].mean(axis=0)
        self.values = self.values - fl
        # self.values[self.values < 0] = 0
        return self

    def riprel(self):
        """
        Replaces drift time coordinate with RIP relative values.
        Useful to cut away the RIP because it´s position is set to 1.

        Does not interpolate the data matrix to a completly artificial
        axis like ims.Dataset.interp_riprel.

        Returns
        -------
        Spectrum
            RIP relative drift time coordinate
            otherwise unchanged.
        """
        rip_index = np.argmax(self.values, axis=1)[0]
        rip_ms = self.drift_time[rip_index]
        dt_riprel = self.drift_time / rip_ms
        self.drift_time = dt_riprel
        self._drift_time_label = 'Drift time RIP relative'
        return self

    def rip_scaling(self):
        """
        Scales values relative to global maximum.
        Can be useful to directly compare spectra from
        instruments with different sensitivity.

        Returns
        -------
        Spectrum
            With scaled values.
        """
        m = np.max(self.values)
        self.values = self.values / m
        return self

    def resample(self, n=2):
        """
        Resamples spectrum by calculating means of every n rows.
        If the length of the retention time is not divisible by n
        it and the data matrix get cropped by the remainder at the long end.

        Parameters
        ----------
        n : int, optional
            Number of rows to mean,
            by default 2.

        Returns
        -------
        Spectrum
            Resampled values.

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> print(sample.shape)
        (4082, 3150)
        >>> sample.resample(2)
        >>> print(sample.shape)
        (2041, 3150)
        """
        a, _ = self.values.shape
        rest = a % n
        if rest != 0:
            self.values = self.values[:a-rest, :]
            self.ret_time = self.ret_time[:a-rest]

        self.values = (self.values[0::n, :] + self.values[1::n, :]) / n
        self.ret_time = self.ret_time[::n]
        return self

    def binning(self, n=2):
        """
        Downsamples spectrum by binning the array with factor n.
        Similar to ims.Spectrum.resampling but works on both dimensions
        simultaneously.
        If the dimensions are not divisible by the binning factor
        shortens it by the remainder at the long end.
        Very effective data reduction because a factor n=2 already 
        reduces the number of features to a quarter.

        Parameters
        ----------
        n : int, optional
            Binning factor, by default 2.

        Returns
        -------
        Spectrum
            Downsampled data matrix.

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> print(sample.shape)
        (4082, 3150)
        >>> sample.binning(2)
        >>> print(sample.shape)
        (2041, 1575)
        """
        a, b = self.values.shape
        rest0 = a % n
        rest1 = b % n

        if rest0 != 0:
            self.values = self.values[:a-rest0, :]
            self.ret_time = self.ret_time[:a-rest0]
        if rest1 != 0:
            self.values = self.values[:, :b-rest1]
            self.drift_time = self.drift_time[:b-rest1]

        new_dims = (a // n, b // n)

        shape = (new_dims[0], a // new_dims[0],
                 new_dims[1], b // new_dims[1])

        self.values = self.values.reshape(shape).mean(axis=(-1, 1))
        self.ret_time = self.ret_time[::n]
        self.drift_time = self.drift_time[::n]
        return self

    def cut_dt(self, start, stop=None):
        """
        Cuts data along drift time coordinate.
        Range in between start and stop is kept.
        If stop is not given uses the end of the array instead.
        Combination with RIP relative drift time
        values makes it easier to cut the RIP away and focus
        on the peak area.

        Parameters
        ----------
        start : int or float
            Start value on drift time coordinate.
        
        stop : int or float, optional
            Stop value on drift time coordinate.
            If None uses the end of the array,
            by default None.

        Returns
        -------
        Spectrum
            New drift time range.

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> print(sample.shape)
        (4082, 3150)
        >>> sample.riprel().cut_dt(1.05, 2)
        >>> print(sample.shape)
        (4082, 1005)
        """
        if stop is None:
            stop = len(self.drift_time)

        idx_start = np.abs(self.drift_time - start).argmin()
        idx_stop = np.abs(self.drift_time - stop).argmin()
        self.drift_time = self.drift_time[idx_start:idx_stop]
        self.values = self.values[:, idx_start:idx_stop]
        return self

    def cut_rt(self, start, stop=None):
        """
        Cuts data along retention time coordinate.
        Range in between start and stop is kept.
        If stop is not given uses the end of the array instead.

        Parameters
        ----------
        start : int or float
            Start value on retention time coordinate.

        stop : int or float, optional
            Stop value on retention time coordinate.
            If None uses the end of the array,
            by default None.

        Returns
        -------
        Spectrum
            New retention time range.

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> print(sample.shape)
        (4082, 3150)
        >>> sample.cut_rt(80, 500)
        >>> print(sample.shape)
        (2857, 3150)
        """
        if stop is None:
            stop = len(self.ret_time)

        idx_start = np.abs(self.ret_time - start).argmin()
        idx_stop = np.abs(self.ret_time - stop).argmin()
        self.ret_time = self.ret_time[idx_start:idx_stop]
        self.values = self.values[idx_start:idx_stop, :]
        return self

    def plot(self, vmin=30, vmax=400, width=9, height=10):
        """
        Plots Spectrum using matplotlibs imshow.
        Use %matplotlib widget in IPython or %matplotlib notebook
        in jupyter notebooks to get an interactive plot widget.
        Returning the figure is needed to make the export plot utilities possible.

        Parameters
        ----------
        vmin : int, optional
            Minimum of color range, by default 30.

        vmax : int, optional
            Maximum of color range, by default 300.

        width : int, optional
            Width in inches, by default 9.

        height : int, optional
            Height in inches, by default 10.

        Returns
        -------
        tuple
            (matplotlib.figure.Figure, matplotlib.pyplot.axes)
        
        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> fig, ax = sample.plot()
        """        

        fig, ax = plt.subplots(figsize=(width, height))

        plt.imshow(
            self.values,
            origin="lower",
            aspect="auto",
            cmap="RdBu_r",
            vmin=vmin,
            vmax=vmax,
            extent=(min(self.drift_time), max(self.drift_time),
                    min(self.ret_time), max(self.ret_time))
            )

        plt.colorbar().set_label("Intensities [arbitrary units]")
        plt.title(self.name)

        ax.xaxis.set_minor_locator(AutoMinorLocator())
        ax.yaxis.set_minor_locator(AutoMinorLocator())

        plt.xlabel(self._drift_time_label)
        plt.ylabel("Retention time [s]")

        return fig, ax

    def export_plot(self, path=None, dpi=300,
                    file_format='jpg', **kwargs):
        """
        Saves the figure as image file. See the docs for
        matplotlib savefig function for supported file formats and kwargs
        (https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.savefig.html).

        Parameters
        ----------
        path : str, optional
            Directory to save the image,
            by default current working directory.

        file_format : str, optional
            See matplotlib savefig docs for information about supported formats,
            by default 'jpg'.

        Example
        -------
        >>> import ims
        >>> sample = ims.Spectrum.read_mea("sample.mea")
        >>> sample.export_plot()
        """
        if path is None:
            path = os.getcwd()

        fig, _ = self.plot(**kwargs)
        fig.savefig(f'{path}/{self.name}.{file_format}', dpi=dpi,
                    bbox_inches="tight", pad_inches=0.2)
